@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright 2011 Stanisław Pitucha. All rights reserved.
+# Copyright 2013 Helmut Grohne. All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification, are
 # permitted provided that the following conditions are met:
@@ -49,6 +50,11 @@ f.read([length])
 random access through seek and tell functions is supported on the archived files
 """
 
+import io
+import struct
+from typing import Optional, List, Dict, BinaryIO, cast
+
+
 HEADER_BSD = 1
 HEADER_GNU = 2
 HEADER_GNU_TABLE = 3
@@ -73,9 +79,8 @@ class ArchiveAccessError(IOError):
 class ArchiveFileHeader(object):
 	""" File header of an archived file, or a special data segment """
 
-	def __init__(self, header, offset):
+	def __init__(self, header: bytes, offset: int):
 		""" Creates a new header from binary data starting at a specified offset """
-		import struct
 
 		name, timestamp, uid, gid, mode, size, magic = struct.unpack(
 				"16s 12s 6s 6s 8s 10s 2s", header)
@@ -113,29 +118,29 @@ class ArchiveFileHeader(object):
 
 		if self.type == HEADER_NORMAL:
 			self.name = name
-			self.file_offset = offset + HEADER_LEN
+			self.file_offset = cast(Optional[int], offset + HEADER_LEN)
 		else:
 			self.name = None
 			self.proxy_name = name
 			self.file_offset = None
 
-	def __repr__(self):
+	def __repr__(self) -> str:
 		""" Creates a human-readable summary of a header """
 		return '''<ArchiveFileHeader: "%s" type:%s size:%i>''' % (self.name,
 				HEADER_TYPES[self.type], self.size)
 
-class ArchiveFileData(object):
+class ArchiveFileData(io.IOBase):
 	""" File-like object used for reading an archived file """
 
-	def __init__(self, file_obj, header):
+	def __init__(self, ar_obj: "Archive", header: ArchiveFileHeader):
 		"""
 		Creates a new proxy for the archived file, reusing the archive's file descriptor
 		"""
 		self.header = header
-		self.file = file_obj
+		self.arobj = ar_obj
 		self.last_offset = 0
 
-	def read(self, size = None):
+	def read(self, size: Optional[int] = None) -> bytes:
 		""" Reads the data from the archived file, simulates file.read """
 		if size is None:
 			size = self.header.size
@@ -143,19 +148,19 @@ class ArchiveFileData(object):
 		if self.header.size < self.last_offset + size:
 			size = self.header.size - self.last_offset
 
-		self.file.seek(self.header.file_offset + self.last_offset)
-		data = self.file.read(size)
+		self.arobj._seek(cast(int, self.header.file_offset) + self.last_offset)
+		data = self.arobj._read(size)
 		if len(data) < size:
 			raise ArchiveAccessError("incorrect archive file")
 
 		self.last_offset += size
 		return data
 
-	def tell(self):
+	def tell(self) -> int:
 		""" Returns the position in archived file, simulates file.tell """
 		return self.last_offset
 
-	def seek(self, offset, whence = 0):
+	def seek(self, offset: int, whence: int = 0) -> int:
 		""" Sets the position in archived file, simulates file.seek """
 		if whence == 0:
 			pass # absolute
@@ -165,38 +170,80 @@ class ArchiveFileData(object):
 			offset += self.header.size
 		else:
 			raise ArchiveAccessError("invalid argument")
-		
+
 		if offset < 0 or offset > self.header.size:
 			raise ArchiveAccessError("incorrect file position")
 		self.last_offset = offset
 
+		return offset
+
+	def seekable(self) -> bool:
+		return self.arobj.seekable
+
 class Archive(object):
 	""" Archive object allowing reading of *.ar files """
-	def __init__(self, filename=None, fileobj=None):
-		self.headers = []
-		self.file = fileobj or open(filename, "rb")
-		if self.file.read(GLOBAL_HEADER_LEN) != b"!<arch>\n":
+
+	def __init__(self, filename: Optional[str] = None, fileobj: Optional[BinaryIO] = None):
+		self.headers = cast(List[ArchiveFileHeader], [])
+		if fileobj:
+			self.file = fileobj
+		elif filename:
+			self.file = open(filename, "rb")
+		else:
+			raise ValueError("either filename or fileobj argument needs to be given")
+		self.position = 0
+		self._detect_seekable()
+		if self._read(GLOBAL_HEADER_LEN) != b"!<arch>\n":
 			raise ArchiveFormatError("file is missing the global header")
-		
+
 		self.next_header_offset = GLOBAL_HEADER_LEN
-		self.gnu_table = None
-		self.archived_files = {}
+		self.gnu_table = cast(Dict[int,bytes], {})
+		self.archived_files = cast(Dict[bytes,ArchiveFileData], {})
 
-	def __read_file_header(self, offset):
+	def _detect_seekable(self) -> None:
+		if hasattr(self.file, 'seekable'):
+			self.seekable = self.file.seekable()
+		else:
+			try:
+				# .tell() will raise an exception as well
+				self.file.tell()
+				self.seekable = True
+			except Exception:
+				self.seekable = False
+
+	def _read(self, length: int) -> bytes:
+		data = self.file.read(length)
+		self.position += len(data)
+		return data
+
+	def _seek(self, offset: int) -> None:
+		if self.seekable:
+			self.file.seek(offset)
+			self.position = self.file.tell()
+		elif offset < self.position:
+			raise ArchiveAccessError("cannot go back when reading archive from a stream")
+		else:
+			# emulate seek
+			while self.position < offset:
+				if not self._read(min(4096, offset - self.position)):
+					# reached EOF before target offset
+					return
+
+	def __read_file_header(self, offset: int) -> Optional[ArchiveFileHeader]:
 		""" Reads and returns a single new file header """
-		self.file.seek(offset)
+		self._seek(offset)
 
-		header = self.file.read(HEADER_LEN)
+		header = self._read(HEADER_LEN)
 
 		if len(header) == 0:
 			return None
 		if len(header) < HEADER_LEN:
 			raise ArchiveFormatError("file header too short")
-		
+
 		file_header = ArchiveFileHeader(header, offset)
 		if file_header.type == HEADER_GNU_TABLE:
 			self.__read_gnu_table(file_header.size)
-			
+
 		add_len = self.__fix_name(file_header)
 		file_header.file_offset = offset + HEADER_LEN + add_len
 
@@ -206,20 +253,20 @@ class Archive(object):
 
 		return file_header
 
-	def __read_gnu_table(self, size):
+	def __read_gnu_table(self, size: int) -> None:
 		""" Reads the table of filenames specific to GNU ar format """
-		table_string = self.file.read(size)
+		table_string = self._read(size)
 		if len(table_string) != size:
 			raise ArchiveFormatError("file too short to fit the names table")
 
 		self.gnu_table = {}
-		
+
 		position = 0
 		for filename in table_string.split(b"\n"):
 			self.gnu_table[position] = filename[:-1] # remove trailing '/'
 			position += len(filename) + 1
 
-	def __fix_name(self, header):
+	def __fix_name(self, header: ArchiveFileHeader) -> int:
 		"""
 		Corrects the long filename using the format-specific method.
 		That means either looking up the name in GNU filename table, or
@@ -234,8 +281,8 @@ class Archive(object):
 			# BSD format includes the filename in the file size
 			header.size -= filename_len
 
-			self.file.seek(header.offset + HEADER_LEN)
-			header.name = self.file.read(filename_len)
+			self._seek(header.offset + HEADER_LEN)
+			header.name = self._read(filename_len)
 			return filename_len
 
 		elif header.type == HEADER_GNU_TABLE:
@@ -246,14 +293,14 @@ class Archive(object):
 			if gnu_position not in self.gnu_table:
 				raise ArchiveFormatError("file references a name not present in the index")
 			header.name = self.gnu_table[gnu_position]
-			
+
 		elif header.type == HEADER_GNU_SYMBOLS:
 			pass
-		
+
 		return 0
 
 	@staticmethod
-	def __pad2(num):
+	def __pad2(num: int) -> int:
 		""" Returns a 2-aligned offset """
 		if num % 2 == 0:
 			return num
@@ -261,12 +308,12 @@ class Archive(object):
 			return num+1
 
 	@staticmethod
-	def __get_bsd_filename_len(name):
+	def __get_bsd_filename_len(name: bytes) -> int:
 		""" Returns the length of the filename for a BSD style header """
 		filename_len = name[3:]
 		return int(filename_len)
 
-	def read_next_header(self):
+	def read_next_header(self) -> Optional[ArchiveFileHeader]:
 		"""
 		Reads a single new header, returning a its representation, or None at the end of file
 		"""
@@ -274,15 +321,27 @@ class Archive(object):
 		if header is not None:
 			self.headers.append(header)
 			if header.type in (HEADER_BSD, HEADER_NORMAL, HEADER_GNU):
-				self.archived_files[header.name] = ArchiveFileData(self.file, header)
+				self.archived_files[header.name] = ArchiveFileData(self, header)
 
 		return header
 
-	def read_all_headers(self):
+	def __next__(self) -> ArchiveFileData:
+		while True:
+			header = self.read_next_header()
+			if header is None:
+				raise StopIteration
+			if header.type in (HEADER_BSD, HEADER_NORMAL, HEADER_GNU):
+				return self.archived_files[header.name]
+	next = __next__
+
+	def __iter__(self) -> "Archive":
+		return self
+
+	def read_all_headers(self) -> None:
 		""" Reads all headers """
 		while self.read_next_header() is not None:
 			pass
 
-	def close(self):
+	def close(self) -> None:
 		""" Closes the archive file descriptor """
 		self.file.close()
